// WacomServerDevice.cpp
/*
	Copyright 2000,2001, 2002  Olaf van Es.   All Rights Reserved.

	Authors:	Olaf van Es			<olaf.van.es@zonnet.nl>
				Frans van Nispen	<frans@xentronix.com>
				Stephan Aßmus		<stippi@yellowbites.com>

*/

#include <fstream.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <Entry.h>
#include <InterfaceDefs.h>
#include <Message.h>
#include <NodeMonitor.h>
#include <OS.h>
#include <Path.h>
#include <Screen.h>
#include <USB.h>
#include <View.h>
#include <File.h>

#include "USBKit.h"
#include "WacomTablet.h"

#include "WacomServerDevice.h"

#define DEFAULT_CLICK_SPEED 250000

#define DEBUG 0

char version[] = "$VER: Wacom USB 2.4 (14 November 2002)";
char copyright[] = "Wacom USB Tablet Driver © 2000,2002 "\
				   "by Olaf van Es, Frans van Nispen & Stephan Aßmus\n";


BInputServerDevice*
instantiate_input_device()
{
	// this is where it all starts
	// make sure this function is exported!

	return (new WacomServerDevice());
}

// constructor
WacomServerDevice::WacomServerDevice()
	: BInputServerDevice(),
	  fTablets(1),
	  fDoubleClickSpeed(DEFAULT_CLICK_SPEED),
	  fActive(false),
	  fLogString("")
{
	if (get_click_speed(&fDoubleClickSpeed) < B_OK)
		fDoubleClickSpeed = DEFAULT_CLICK_SPEED;
}

// destructor
WacomServerDevice::~WacomServerDevice()
{
	// cleanup
	_StopAll();
	while (WacomTablet* tablet = (WacomTablet*)fTablets.RemoveItem(0L))
		delete tablet;
}

// InitCheck
status_t
WacomServerDevice::InitCheck()
{
	input_device_ref wacomDevice = { "Wacom USB Tablet",
									 B_POINTING_DEVICE, (void*)this };
	input_device_ref* wacomDeviceList[2] = { &wacomDevice, NULL };
	RegisterDevices(wacomDeviceList);

	return B_OK;
}

// SystemShuttingDown
status_t
WacomServerDevice::SystemShuttingDown()
{
	// do any cleanup (ie. saving a settings file) when the
	// system is about to shut down

	_StopAll();

#if DEBUG
	DumpLogString("/boot/home/wacom.log");
#endif

	return (BInputServerDevice::SystemShuttingDown());
}

// Start
status_t
WacomServerDevice::Start(const char* device, void* cookie)
{
	// start generating events
	// this is a hook function, it is called for you
	// (you should not call it yourself)

	fActive = true;
	
	for (int32 i = 0; WacomTablet* tablet = (WacomTablet*)fTablets.ItemAt(i); i++)
#if DEBUG
	{
		status_t status = tablet->Start();
		fLogString << "started tablet device: " << strerror(status) << "\n";
	}
#else
		tablet->Start();
#endif

	USBRoster::Start();

	return B_OK;
}

// Stop
status_t
WacomServerDevice::Stop(const char* device, void* cookie)
{
	// stop generating events
	// this is a hook function, it is called for you
	// (you should not call it yourself)

	fActive = false;

	_StopAll();

	USBRoster::Stop();

	return B_OK;
}

// Control
status_t
WacomServerDevice::Control(const char* device, void* cookie, uint32 code, BMessage* message)
{
	// respond to changes in the system
	// this is a hook function, it is called for you
	// (you should not call it yourself)

	switch (code) {
		case B_CLICK_SPEED_CHANGED:
			if (get_click_speed(&fDoubleClickSpeed) < B_OK)
				fDoubleClickSpeed = DEFAULT_CLICK_SPEED;
			break;
		default:
			BInputServerDevice::Control(device, cookie, code, message);
			break;
	}
	return B_OK;
}

// DeviceAdded
//
// A new device was attached to the USB chain.
// Check if it's a supported device, if so, create a new WacomTablet
// object and start it.
status_t
WacomServerDevice::DeviceAdded(USBDevice* device)
{
	status_t status = B_BAD_VALUE;
	// Check if it's a supported device...
	WacomTablet* tablet = new WacomTablet(this);
	if (tablet->SetTo(device) >= B_OK) {
		// add tablet to our list
		if (fTablets.AddItem((void*)tablet)) {
#if DEBUG
			fLogString << "wacom tablet added\n";
#endif
			// start tablet polling only if we're started
			if (fActive)
				tablet->Start();
			status = B_OK;
		} else {
			delete tablet; // will delete the device as well
			status = B_NO_MEMORY;
		}
	} else {
#if DEBUG
		fLogString << "device not added\n";
#endif
		delete tablet;
	}
	return status;
}


// DeviceRemoved
//
// A Device has been removed from the USB chain.
// Check if it's one of ours, if so, stop the device, then delete it.
void
WacomServerDevice::DeviceRemoved(USBDevice* device)
{
	for (int32 i = 0; WacomTablet* tablet = (WacomTablet*)fTablets.ItemAt(i); i++) {
		if (tablet->Device() == device) {
#if DEBUG
		fLogString << "wacom tablet removed\n";
#endif
			if (fTablets.RemoveItem((void*)tablet))
				delete tablet;
			break;
		}
	}
}

// LogDataBytes
void
WacomServerDevice::LogDataBytes(uchar* data, int bytes)
{
	for (int32 i = 0; i < bytes; i += 2)
		fLogString << (uint32)data[i] << " " << (uint32)data[i + 1] << "  ";
	fLogString << "\n";
}

// DumpLogString
void
WacomServerDevice::DumpLogString(const char* path)
{
	if (fLogString.Length() > 0) {
		BFile logFile(path, B_WRITE_ONLY | B_ERASE_FILE | B_CREATE_FILE);
		if (logFile.InitCheck() >= B_OK) {
			logFile.Write(fLogString.String(), fLogString.Length());
			fLogString.SetTo("");
		}
		logFile.Unset();
	}
}

// _StopAll
void
WacomServerDevice::_StopAll()
{
	for (int32 i = 0; WacomTablet* tablet = (WacomTablet*)fTablets.ItemAt(i); i++)
		tablet->Stop();
}

