// WacomTablet.cpp
/*
	Copyright 2000,2001, 2002  Olaf van Es.   All Rights Reserved.

	Authors:	Olaf van Es			<olaf.van.es@zonnet.nl>
				Frans van Nispen	<frans@xentronix.com>
				Stephan Aßmus		<stippi@yellowbites.com>
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <InterfaceDefs.h>
#include <Screen.h>
#include <View.h>

#include "USBKit.h"
#include "WacomServerDevice.h"

#include "WacomTablet.h"

#define SNOOZE_AMOUNT 2500
#define JITTER_X 8164800.0
#define JITTER_Y 4444311.0
#define ACCELERATION_KICK_IN 2.3

#define DEBUG 0

// constructor
WacomTablet::WacomTablet(WacomServerDevice* parent)
	: fParent(parent),
	  fActive(false),
	  fThreadID(B_ERROR),
	  fDevice(NULL),
	  fDeviceMode(DEVICE_UNKOWN),
	  fMaxX(1.0),
	  fMaxY(1.0),
	  fDataBytes(10),
	  fPosX(0.5),
	  fPosY(0.5),
	  fFakeMouseX(0.5),
	  fFakeMouseY(0.5),
	  fButtons(0),
	  fPressure(0.0),
	  fModifiers(0),
	  fEraser(0),
	  fTiltX(0.0),
	  fTiltY(0.0),
	  fClicks(0),
	  fHasContact(false)
{
}

// destructor
WacomTablet::~WacomTablet()
{
	// cleanup
	Stop();
}

// SetTo
status_t
WacomTablet::SetTo(USBDevice* device)
{
	status_t status = B_BAD_VALUE;
	if (device) {
		status = DetectDevice(*device);
		if (status >= B_OK) {
			status = InitDevice(*device);
			if (status >= B_OK) {
				delete fDevice;
				fDevice = device;
			}
		}
	}
	return status;
}

// Start
status_t
WacomTablet::Start()
{
	status_t status = B_NO_INIT;
	if (fDevice) {
		fActive = true;
		// get a nice name for our polling thread
		char* name;
		_GetName(*fDevice, &name);
		// start generating events
		fThreadID = spawn_thread(poll_usb_device, name, 104, this);
		if (fThreadID >= B_OK) {
			resume_thread(fThreadID);
			status = B_OK;
		} else
			status = fThreadID;
	}
	return status;
}

// Stop
void
WacomTablet::Stop()
{
//	status_t err = B_OK;

	fActive = false;
	if (fThreadID >= B_OK)
		// I couldn't figure this out.
		// Appearantly, when the polling thread happens
		// to block on the usb interrupt transfer, we don't
		// exit nicely. My tablet for example doesn't transfer
		// when neither pen nor mouse are on the tablet, so the
		// transfer request blocks indefinately.
		// however, kill_thread() works every single time,
		// so what the heck.... :-/
//		wait_for_thread(fThreadID, &err);
		kill_thread(fThreadID);
	fThreadID = B_ERROR;
}

// DetectDevice
status_t
WacomTablet::DetectDevice(USBDevice& device)
{
	status_t status = device.InitCheck();
	if (status == B_OK) {
		status = B_BAD_VALUE;
		if (device.VendorID() == 0x056a) { // wacom
			status = B_OK;
			switch (device.ProductID()) {
				case 0x00:
					SetDevice(5040.0, 3780.0, DEVICE_PENPARTNER, 7);
					break;
				case 0x10:
				case 0x11:
				case 0x12:
					SetDevice(10206.0, 7422.0, DEVICE_GRAPHIRE, 8);
					break;
				case 0x20:
					SetDevice(12700.0, 10360.0);
					break;
				case 0x21:
					SetDevice(20320.0, 15040.0);
					break;
				case 0x22:
					SetDevice(30480.0, 23060.0);
					break;
				case 0x23:
					SetDevice(30480.0, 30480.0);
					break;
				case 0x24:
					SetDevice(47720.0, 30480.0);
					break;
				case 0x30:
				case 0x31:
				case 0x32:
				case 0x33:
				case 0x34:
				case 0x35:
					SetDevice(12328.0, 9256.0, DEVICE_PL500, 8);
				case 0x41:
					SetDevice(12700.0, 10360.0);
					break;
				case 0x42:
					SetDevice(20320.0, 15040.0);
					break;
				case 0x43:
					SetDevice(30480.0, 23060.0);
					break;
				case 0x44:
					SetDevice(30480.0, 30480.0);
					break;
				case 0x45:
					SetDevice(47720.0, 30480.0);
					break;
				default:
					status = B_ERROR;
					break;
			}
		}
	}
/*if (status >= B_OK)
*fParent->LogString() << "wacom device (class: "
					  << device.Class()	
					  << ", subclass: "
					  << device.Subclass()
					  << ", protocol: "
					  << device.Protocol() << ")\n";
fParent->DumpLogString("/boot/home/wacom.log");*/
	return status;
}

// SetDevice
void
WacomTablet::SetDevice(float maxX, float maxY, uint32 mode,  int dataBytes)
{
	fDeviceMode = mode;
	fMaxX = maxX;
	fMaxY = maxY;
	fDataBytes = dataBytes;
	fJitterX = fMaxX / JITTER_X;
	fJitterY = fMaxY / JITTER_Y;
}

// InitDevice
status_t
WacomTablet::InitDevice(USBDevice& device)
{
	status_t status = device.InitCheck();
	if (status >= B_OK) {
		status = device.SetConfiguration(device.ConfigurationAt(0));
		if (status >= B_OK) {
			// you won't believe the amount of time it took to figure this one out.			
			device.ControlTransfer(0x21, 0x0b, 1, 0, 0, 0);

			// set Interface
			char repData[2] = { 0x02, 0x02 };
			device.ControlTransfer(0x21, 0x09, (3 << 8) + 2, 1, 2, repData);		
		}
	}
	return status;
}

// ReadData
void
WacomTablet::ReadData(uchar* data, bool& hasContact, uint32& mode,
						 uint32& buttons, float& x, float& y, float& pressure,
						 int32& clicks, int32& eraser,
						 float& wheelX, float& wheelY,
						 float& tiltX, float& tiltY) const
{
	hasContact = false;
	buttons = 0;
	mode = MODE_PEN;
	bool firstButton = false;
	bool secondButton = false;
	bool thirdButton = false;
	uint16 xPos = 0;
	uint16 yPos = 0;

	switch (fDeviceMode) {
		case DEVICE_PENPARTNER: {
			xPos = data[2] << 8 | data[1];
			yPos = data[4] << 8 | data[3];

			eraser = (data[5] & 0x20);

			int8 pressureData = data[6];
			pressure = (float)(pressureData + 120) / 240.0;

			firstButton = ((pressureData > -80) && !(data[5] & 0x20));
			secondButton = (data[5] & 0x40);

			hasContact = true;
			break;
		}
		case DEVICE_GRAPHIRE: {
			xPos = data[3] << 8 | data[2];
			yPos = data[5] << 8 | data[4];

			hasContact = (data[1] & 0x80);		

			uint16 pressureData = data[7] << 8 | data[6];
			pressure = (float)pressureData / 511.0;
			eraser = (data[1] & 0x20);

			// mouse wheel support
			if (data[1] & 0x40) {	// mouse is on tablet!
#if DEBUG
fParent->LogDataBytes(data);
#endif
				wheelY = (float)(int8)data[6];
				mode = MODE_MOUSE;
				// override contact to loose it as soon as possible
				// when mouse is lifted from tablet
				hasContact = (uint8)data[7] >= 30;
				pressure = 0.0;
				eraser = 0;
			}

			firstButton = pressure > 0.0 ? true : (data[1] & 0x01);
			secondButton = (data[1] & 0x02);
			thirdButton = (data[1] & 0x04);

			break;
		}
		case DEVICE_INTUOS:
			if ((data[0] == 0x02) && !(((data[1] >> 5) & 0x03) == 0x02)) {
				xPos = data[2] << 8 | data[3];
				yPos = data[4] << 8 | data[5];
				uint16 pressureData = data[6] << 2 | ((data[7] >> 6) & 0x03);
				pressure = (float)pressureData / 1023.0;
				// eraser = (data[1] & 0x20);  eraser is een tool-id
				firstButton = (pressureData > 180);
				secondButton = (data[1] & 0x02);
				thirdButton = (data[1] & 0x04);
				hasContact = (data[1] & 0x40);			
				// convert tilt (-128 ... 127)
				int8 tiltDataX = ((data[7] & 0x3f) << 2) | ((data[8] & 0x80) >> 6);
				int8 tiltDataY = (data[8] & 0x7f) << 1;
				// convert to floats
				tiltX = (float)tiltDataX / 128.0;
				tiltY = (float)tiltDataY / 128.0; 		
			}
			break;
		case DEVICE_PL500: {
			hasContact = ( data[1] & 0x20);
			xPos = data[2] << 8 | data[3];
			yPos = data[5] << 8 | data[6];
			firstButton = (data[4] & 0x08);				
			secondButton = (data[4] & 0x10);
			thirdButton = (data[4] & 0x20);
			uint16 pressureData = (data[4] & 0x04) >> 2 | (data[7] & 0x7f) << 1;
			pressure = (float)pressureData / 511.0;
			break;
		}
	}
	if (pressure > 1.0)
		pressure = 1.0;
	else if (pressure < 0.0)
		pressure = 0.0;
	buttons = (firstButton ? B_PRIMARY_MOUSE_BUTTON : 0)
			   | (secondButton ? B_SECONDARY_MOUSE_BUTTON : 0)
			   | (thirdButton ? B_TERTIARY_MOUSE_BUTTON : 0);
	x = (float)xPos;
	y = (float)yPos;
}

// SetStatus
void
WacomTablet::SetStatus(uint32 mode, uint32 buttons, float x, float y, float pressure,
						  int32 clicks, uint32 modifiers, int32 eraser,
						  float wheelX, float wheelY, float tiltX, float tiltY)
{
	if (fActive) {
		uint32 what = B_MOUSE_MOVED;
		if (buttons > fButtons)
			what = B_MOUSE_DOWN;
		else if (buttons < fButtons)
			what = B_MOUSE_UP;
	
		x /= fMaxX;
		y /= fMaxY;
	
		float deltaX = 0.0;
		float deltaY = 0.0;
	
		float absDeltaX = 0.0;
		float absDeltaY = 0.0;

		float unfilteredX = x;
		float unfilteredY = y;

		if (fHasContact) {
			deltaX = x - fPosX;
			deltaY = y - fPosY;

			absDeltaX = fabsf(deltaX);
			absDeltaY = fabsf(deltaY);

#if DEBUG
*fParent->LogString() << "x: " << x << ", y: " << y << ", pressure: " << pressure << "\n\n";
#endif
			// apply a bit of filtering
			if (absDeltaX < fJitterX)
				x = fPosX;
			if (absDeltaY < fJitterY)
				y = fPosY;
		}
	
		// only do send message if something changed
		if (x != fPosX || y != fPosY || fButtons != buttons || pressure != fPressure
			|| fEraser != eraser || fTiltX != tiltX || fTiltY != tiltY) {
	
			bigtime_t now = system_time();
	
			// common fields for any mouse message
			BMessage* event = new BMessage(what);
			event->AddInt64("when", now);
			event->AddInt32("buttons", buttons);
			if (mode == MODE_PEN) {
				event->AddFloat("x", unfilteredX);
				event->AddFloat("y", unfilteredY);
				event->AddFloat("be:tablet_x", unfilteredX);
				event->AddFloat("be:tablet_y", unfilteredY);
				event->AddFloat("be:tablet_pressure", pressure);
				event->AddInt32("be:tablet_eraser", eraser);
				if (fDeviceMode == DEVICE_INTUOS) {
					event->AddFloat("be:tablet_tilt_x", tiltX);
					event->AddFloat("be:tablet_tilt_y", tiltY);
				}
				// adjust mouse coordinates as well
				// to have the mouse appear at the pens
				// last position when switching
				fFakeMouseX = unfilteredX;
				fFakeMouseY = unfilteredY;
			} else if (mode == MODE_MOUSE) {
				// apply acceleration
				float accelerationX = fJitterX * ACCELERATION_KICK_IN;
//				if (absDeltaX > accelerationX)
					deltaX *= absDeltaX / accelerationX;
				float accelerationY = fJitterY * ACCELERATION_KICK_IN;
//				if (absDeltaY > accelerationY)
					deltaY *= absDeltaY / accelerationY;
				// calculate screen coordinates
				fFakeMouseX = min_c(1.0, max_c(0.0, fFakeMouseX + deltaX));
				fFakeMouseY = min_c(1.0, max_c(0.0, fFakeMouseY + deltaY));
				event->AddFloat("x", fFakeMouseX);
				event->AddFloat("y", fFakeMouseY);
				event->AddFloat("be:tablet_x", fFakeMouseX);
				event->AddFloat("be:tablet_y", fFakeMouseY);
			}
			event->AddInt32("modifiers", modifiers);
	
			// additional fields for mouse down or up
			if (what == B_MOUSE_DOWN) {
				if (now - fLastClickTime < fParent->DoubleClickSpeed())
					fClicks++;
				else
					fClicks = 1;
				fLastClickTime = now;
				event->AddInt32("clicks", fClicks);
			} else if (what == B_MOUSE_UP)
				event->AddInt32("clicks", 0);
										
			fParent->EnqueueMessage(event);
	
			// apply values to members
			fPosX = x;
			fPosY = y;
			fButtons = buttons;
			fPressure = pressure;
			fModifiers = modifiers;
			fEraser = eraser;
			fTiltX = tiltX;
			fTiltY = tiltY;
		}
	
		// separate wheel changed message
		if (fWheelX != wheelX || fWheelY != wheelY) {
			BMessage* event = new BMessage(B_MOUSE_WHEEL_CHANGED);
			event->AddInt64("when", system_time());
			event->AddFloat("be:wheel_delta_x", wheelX);
			event->AddFloat("be:wheel_delta_y", wheelY);
			fParent->EnqueueMessage(event);
	
			// apply values to members
			fWheelX = wheelX;
			fWheelY = wheelY;
		}
	}
}

// SetContact
void
WacomTablet::SetContact(bool contact)
{
	fHasContact = contact;
}

// poll_usb_device
int32
WacomTablet::poll_usb_device(void* arg)
{
	WacomTablet *wacomDevice = (WacomTablet*)arg;

	uchar data[12];

	USBDevice* device = wacomDevice->Device();

	if (device) {
		const USBEndpoint* endPoint = device->ActiveConfiguration()->InterfaceAt(0)->EndpointAt(0);
		int dataBytes = wacomDevice->DataBytes();
		while (wacomDevice->IsActive()) {
			if (endPoint->InterruptTransfer(data, dataBytes) == dataBytes) {
				// data we read from the wacom device
				uint32 mode;
				bool hasContact = false;
				uint32 buttons = 0;
				float x = 0.0;
				float y = 0.0;
				float pressure = 0.0;
				int32 clicks = 0;
				int32 eraser = 0;
				float wheelX = 0.0;
				float wheelY = 0.0;
				float tiltX = 0.0;
				float tiltY = 0.0;
				// let the device extract all information from the data
				wacomDevice->ReadData(data, hasContact, mode, buttons,
									  x, y, pressure, clicks, eraser,
									  wheelX, wheelY, tiltX, tiltY);
				if (hasContact) {
					// apply the changes to the device
					wacomDevice->SetStatus(mode, buttons, x, y, pressure,
										   clicks, modifiers(), eraser,
										   wheelX, wheelY, tiltX, tiltY);
				}
				wacomDevice->SetContact(hasContact);
				snooze(SNOOZE_AMOUNT);
			} else
				snooze(SNOOZE_AMOUNT);
		}
	}
	return B_OK;
}

// _GetName
void
WacomTablet::_GetName(USBDevice& device, char** name) const
{
	if (device.VendorID()==0x056a) {
		switch (device.ProductID()) {
			case 0x00:
				*name = "Wacom USB";
				break;
			case 0x10:
				*name = "Wacom Graphire USB";
				break;
			case 0x11:
				*name = "Wacom Graphire2 4x5\" USB";
				break;
			case 0x12:
				*name = "Wacom Graphire2 5x7\" USB";
				break;
			case 0x20:
				*name = "Wacom Intuos 4x5\" USB";
				break;
			case 0x21:
				*name = "Wacom Intuos 6x8\" USB";
				break;
			case 0x22:
				*name = "Wacom Intuos 9x12\" USB";
				break;
			case 0x23:
				*name = "Wacom Intuos 12x12\" USB";
				break;
			case 0x24:
				*name = "Wacom Intuos 12x18\" USB";
				break;
			case 0x30:
				*name = "Wacom PL400 USB";
				break;
			case 0x31:
				*name = "Wacom PL500 USB";
				break;
			case 0x32:
				*name = "Wacom PL600 USB";
				break;
			case 0x33:
				*name = "Wacom PL600SX USB";
				break;
			case 0x34:
				*name = "Wacom PL550 USB";
				break;
			case 0x35:
				*name = "Wacom PL800 USB";
				break;
			case 0x41:
				*name = "Wacom Intuos2 4x5\" USB";
				break;
			case 0x42:
				*name = "Wacom Intuos2 6x8\" USB";
				break;
			case 0x43:
				*name = "Wacom Intuos2 9x12\" USB";
				break;
			case 0x44:
				*name = "Wacom Intuos2 12x12\" USB";
				break;
			case 0x45:
				*name = "Wacom Intuos2 12x18\" USB";
				break;
		}
	}
}
