// 	Copyright 2000 Be Incorporated. All Rights Reserved.
//	This file may be used under the terms of the Be Sample Code 
//	License.

#include <OS.h>
#include <Drivers.h>
#include <KernelExport.h>
#include <stdlib.h>
#include <stdio.h>

#include <USB.h>
//#include "audio.h"

#define DEBUG_DRIVER 1

#if DEBUG_DRIVER
#define DPRINTF(x) dprintf x
#else
#define DPRINTF(x) ((void)0)
#endif

typedef struct wacom_device wacom_device;

struct wacom_device {
	wacom_device*		next;

	int					open;
	int					number;
	const usb_device*	dev;
};

// handy strings for referring to ourself
#define ID "wacom: "
static const char *drivername = "wacom";
static const char *basename = "input/wacom_tablet/";

// list of device instances and names for publishing
static wacom_device *device_list = NULL;
static sem_id dev_list_lock = -1;
static int device_count = 0;

static char **device_names = NULL;

// handles for the USB bus manager
static char *usb_name = B_USB_MODULE_NAME;
static usb_module_info *usb;

// These rather inelegant routines are used to assign numbers to
// device instances so that they have unique names in devfs.

static uint32 device_numbers = 0;

static int 
get_number()
{
	int num;
	
	for(num = 0; num < 32; num++) {
		if(!(device_numbers & (1 << num))){
			device_numbers |= (1 << num);
			return num;
		}
	}
	
	return -1;
}

static void
put_number(int num)
{
	device_numbers &= ~(1 << num);
}

// Device addition and removal ---------------------------------------
//
// add_device() and remove_device() are used to create and tear down
// device instances.  They are driver by the callbacks device_added()
// and device_removed() which are invoked by the USB bus manager.

static wacom_device* 
add_device(const usb_device *dev, const usb_configuration_info *conf)
{
	wacom_device *ad = NULL;
	int num, i, ifc, alt;
	const usb_interface_info *ii;
	status_t st;
	usb_pipe *out;
	const usb_device_descriptor* udd;
	
	DPRINTF((ID "add_device(%p, %p)\n", dev, conf));
	
	if ((num = get_number()) < 0)
		return NULL;

	udd = usb->get_device_descriptor(dev);
	if (udd && udd->vendor_id == 0x056a) {

		DPRINTF((ID "add_device() - wacom tablet detected\n"));

		for (ifc = 0; ifc < conf->interface_count; ifc++) {
			DPRINTF((ID "add_device() - examining interface: %d\n", ifc));
			for (alt = 0; alt < conf->interface[ifc].alt_count; alt++) {
				ii = &conf->interface[ifc].alt[alt];

				DPRINTF((ID "add_device() - examining alt interface: %d\n", alt));

				// only accept input endpoints
				if (ii->endpoint[0].descr->endpoint_address & 0x80) {
					// does it support 2 channel, 16 bit audio?
/*					for(i = 0; i < ii->generic_count; i++){
						usb_format_type_descr *ft = (usb_format_type_descr*) ii->generic[i];
						if(ft->type != AC_CS_INTERFACE) continue;
						if(ft->subtype != AC_FORMAT_TYPE) continue;
						if(ft->length < sizeof(usb_format_type_descr)) continue;
						if(ft->num_channels != 2) continue;
						if(ft->subframe_size != 2) continue;
						if(ft->sample_freq_type != 0) continue;
						dprintf(ID "found a wacom tablet (ifc=%d, alt=%d, mp=%d)\n",
								ifc, alt, ii->endpoint[0].descr->max_packet_size);
		
						goto got_one;
					}*/
					DPRINTF((ID "add_device() - found input endpoint\n"));
					goto got_one;
				}
			}
		}
	} else
		goto fail;

fail:
	put_number(num);
	if (ad)
		free(ad);
	return NULL;

got_one:
	if ((ad = (wacom_device *) malloc(sizeof(wacom_device))) == NULL)
		goto fail;
	
	ad->dev = dev;
	ad->number = num;
	ad->open = 0;
	
	dprintf(ID "add_device() - setting interface: %p", ii);
	if ((st = usb->set_alt_interface(dev, ii)) != B_OK) {
		dprintf(ID "\n -> set_alt_interface(0) returns %d\n", st);
		goto fail;
	} else
		dprintf(ID " ... success!\n");

	dprintf(ID "add_device() - setting configuration");
	if ((st = usb->set_configuration(dev, conf)) != B_OK) {
		dprintf(ID "\n -> set_configuration() returns %d\n", st);
		goto fail;
	} else
		dprintf(ID " ... success!\n");
	
//	out = ii->endpoint[0].handle;
//	DPRINTF((ID "added %p (out=%p) (/dev/%s%d)\n", ad, out, basename, num));
DPRINTF((ID "add_device() - added %p (/dev/%s%d)\n", ad, basename, num));
	
/*	if((st = usb->set_pipe_policy(out, NB, TS, SAMPLE_SIZE)) != B_OK){
		dprintf(ID "set_pipe_policy(out) returns %d\n", st);
		goto fail;
	}

	init_iso_channel(&ad->out_channel, out, qs_out, FALSE);
	start_iso_channel_out(&ad->out_channel);
*/
	// add it to the list of devices so it will be published, etc
	acquire_sem(dev_list_lock);
	ad->next = device_list;
	device_list = ad;
	device_count++;
	release_sem(dev_list_lock);
	
	return ad;
}

static void 
remove_device(wacom_device *ad)
{
/*	if(ad->out_channel.active) stop_iso_channel(&ad->out_channel);
	uninit_iso_channel(&ad->out_channel);*/
	put_number(ad->number);
	free(ad);
}

static status_t 
device_added(const usb_device *dev, void **cookie)
{
	const usb_configuration_info *conf;
	wacom_device *ad;
	int i;

	DPRINTF((ID "device_added(%p,...)\n", dev));

	if ((conf = usb->get_nth_configuration(dev, 0)) != NULL) {
		if ((ad = add_device(dev, conf)) != NULL) {
			*cookie = (void*)ad;
			DPRINTF((ID "device_added() - done!\n"));
			return B_OK;
		} else
			DPRINTF((ID "device_added() - failed to add device!\n"));
	} else
		DPRINTF((ID "device_added() - failed to get configuration!\n"));

	return B_ERROR;
}


static status_t 
device_removed(void *cookie)
{
	wacom_device *ad = (wacom_device *) cookie;
	int i;

	DPRINTF((ID "device_removed(%p)\n", ad));

	if (ad) {
		
		acquire_sem(dev_list_lock);
	
		// mark it as inactive and encourage IO to finish
//		ad->out_channel.active = 0;
//		delete_sem(ad->out_channel.num_available_packets);
	
		// remove it from the list of devices
		if (ad == device_list) {
			device_list = ad->next;
		} else {
			wacom_device *n;
			for (n = device_list; n; n = n->next) {
				if (n->next == ad) {
					n->next = ad->next;
					break;
				}
			}
		}
		device_count--;
		
		// tear it down if it's not open -- 
		// otherwise the last device_free() will handle it
			
		if (ad->open == 0) {
			remove_device(ad);
		} else {
			DPRINTF((ID "device /dev/%s%d still open -- marked for removal\n",
					 basename, ad->number));
		}
		
		release_sem(dev_list_lock);
	}

	return B_OK;
}

// Device Hooks -----------------------------------------------------------
//
// Here we implement the posixy driver hooks (open/close/read/write/ioctl)

static status_t
device_open(const char *dname, uint32 flags, void **cookie)
{	
	wacom_device *ad;
	int n;

	n = atoi(dname + strlen(basename)); 

	DPRINTF((ID "device_open(\"%s\",%d,...)\n", dname, flags));

	acquire_sem(dev_list_lock);
	for (ad = device_list; ad; ad = ad->next) {
		if (ad->number == n) {
//			if(ad->out_channel.active) {
				ad->open++;
				*cookie = ad;	
				release_sem(dev_list_lock);
				return B_OK;
//			} else {
//				dprintf(ID "device is going offline. cannot open\n");
//			}
		}
	}
	release_sem(dev_list_lock);
	return B_ERROR;
}
 
static status_t
device_close (void *cookie)
{
	wacom_device *ad = (wacom_device*) cookie;
	if (ad) {
		DPRINTF((ID "device_close() name = \"%s%d\"\n", basename, ad->number));
		return B_OK;
	}
	return B_ERROR;
}

static status_t
device_free(void *cookie)
{
	wacom_device *ad = (wacom_device *)cookie;
	
	DPRINTF((ID "device_free() name = \"%s%d\"\n", basename, ad->number));
	
	acquire_sem(dev_list_lock);
	ad->open--;
	if ((ad->open == 0)/* && (ad->out_channel.active == 0)*/)
		remove_device(ad);
	release_sem(dev_list_lock);

	return B_OK;
}

static status_t
device_read(void *cookie, off_t pos, void *buf, size_t *count)
{
	return B_ERROR;
}

static status_t
device_write(void *cookie, off_t pos, const void *buf, size_t *count)
{
/*	wacom_device* ad = (wacom_device*) cookie; 
	iso_channel* channel = &ad->out_channel;
	status_t st;

#if 0
	DPRINTF((ID "device_write(%p,%Ld,0x%x,%d) name = \"%s%d\"\n",
			 cookie, pos, buf, *count, basename, ad->number));
#endif
		
	if(channel->remain == 0) {
		st = acquire_sem_etc(channel->num_available_packets, 1, B_RELATIVE_TIMEOUT, 4*1000*1000);
		if(st) {
			if(st == B_TIMED_OUT) {
				dprintf("st = B_TIMED_OUT\n");
				*count = 0;
			}
			return st;
		}
	
		if(channel->current_rw == channel->next_to_queue) dprintf("e");
		
		channel->remain = channel->current_rw->buffer_size;
		channel->buf_ptr = channel->current_rw->buffer;
	}
	
	if(channel->remain >= *count) {
		memcpy(channel->buf_ptr, buf, *count);
		channel->remain -= *count;
		channel->buf_ptr += *count;
	} else {
		memcpy(channel->buf_ptr, buf, channel->remain);
		*count = channel->remain;
		channel->remain = 0;
		channel->buf_ptr = NULL;
	}

	return B_OK;
*/	
return B_ERROR;
}

static status_t
device_control (void *cookie, uint32 msg, void *arg1, size_t len)
{
	return B_ERROR;
}

// Driver Hooks ---------------------------------------------------------
//
// These functions provide the glue used by DevFS to load/unload
// the driver and also handle registering with the USB bus manager
// to receive device added and removed events

static usb_notify_hooks notify_hooks = 
{
	&device_added,
	&device_removed
};

usb_support_descriptor supported_devices[] =
{
	{ 0, 0, 0, 0x056a, 0 }
};

_EXPORT status_t 
init_hardware(void)
{
	return B_OK;
}

_EXPORT status_t
init_driver(void)
{
	int i;
	DPRINTF((ID "init_driver(), built %s %s\n", __DATE__, __TIME__));

#if DEBUG_DRIVER
	if (load_driver_symbols(drivername) == B_OK) {
		DPRINTF((ID "loaded symbols\n"));
	} else {
		DPRINTF((ID "no symbols for you!\n"));
	}
#endif

	if (get_module(usb_name,(module_info**) &usb) != B_OK) {
		dprintf(ID "cannot get module \"%s\"\n",usb_name);
		return B_ERROR;
	} 

	if ((dev_list_lock = create_sem(1,"dev_list_lock")) < 0) {
		put_module(usb_name);
		return dev_list_lock;
	}

	usb->register_driver(drivername, supported_devices, 1, NULL);
	usb->install_notify(drivername, &notify_hooks);

	return B_OK;
}

_EXPORT void
uninit_driver(void)
{
	int i;
	
	DPRINTF((ID "uninit_driver()\n"));
	
	usb->uninstall_notify(drivername);
	
	delete_sem(dev_list_lock);

	put_module(usb_name);

	if (device_names) {
		for (i = 0; device_names[i]; i++)
			free(device_names[i]);
		free(device_names);
	}
	
}

_EXPORT const char**
publish_devices()
{
	wacom_device *ad;
	int i;
	
	DPRINTF((ID "publish_devices()\n"));
	
	if (device_names) {
		for (i = 0; device_names[i]; i++)
			free((char *) device_names[i]);
		free(device_names);
	}
	
	acquire_sem(dev_list_lock);	
	device_names = (char **) malloc(sizeof(char*) * (device_count + 1));
	if (device_names) {
		for (i = 0, ad = device_list; ad; ad = ad->next) {
//			if (ad->out_channel.active) {
				if (device_names[i] = (char *) malloc(strlen(basename) + 4)) {
					sprintf(device_names[i],"%s%d",basename,ad->number);
					DPRINTF((ID "publishing: \"/dev/%s\"\n",device_names[i]));
					i++;
				}
//			}
		}
		device_names[i] = NULL;
	}

	release_sem(dev_list_lock);
	
	return (const char **) device_names;
}

static device_hooks DeviceHooks = {
	device_open, 			
	device_close, 			
	device_free,			
	device_control, 		
	device_read,			
	device_write			 
};

_EXPORT device_hooks*
find_device(const char* name)
{
	return &DeviceHooks;
}
